/* 
 * This file cannot include wrappers.h because it needs original
 * versions of C library and sys call functions
 */

#include <stdio.h>
#include <malloc.h>
#include <stdlib.h>
#include <dirent.h>
#include "spinlock.h"

int mem_usage=0;
int max_mem_usage=0;

AT_SPINLOCK_DEC(kernel_lck);  /* Serializes kernel calls */

/* For now, serialize all allocation */
void *at_malloc (long size)
{
  void *sto;

  AT_KERNEL_LOCK;
  sto = malloc (size);
  mem_usage+=size;
  if (mem_usage > max_mem_usage) max_mem_usage = mem_usage;
  AT_KERNEL_UNLOCK;
  if (!sto) {
    perror ("at_malloc");
    exit (1);
  }
  return (sto);
}

void *at_calloc(long nelem, long elsize){
  void *s;

  AT_KERNEL_LOCK;
  s = malloc(nelem*elsize);
  mem_usage+=nelem*elsize;
  if (mem_usage > max_mem_usage) max_mem_usage = mem_usage;  
  AT_KERNEL_UNLOCK;
  if (!s) {
    perror("at_calloc");
    exit (1);
  }
  return (s);
}


void *at_realloc(void *a, long size){
  void *s;

  AT_KERNEL_LOCK;
  s = realloc(a, size);
  mem_usage+=size;
  if (mem_usage > max_mem_usage) max_mem_usage = mem_usage;  
  AT_KERNEL_UNLOCK;
  if (!s) {
    perror("at_realloc");
    exit (1);
  }
  return (s);
}



void at_free(void* p){
  AT_KERNEL_LOCK;
  free(p);
  AT_KERNEL_UNLOCK;
}

extern void *alloca(size_t size);

void *at_alloca(long size){
  void *s;

  AT_KERNEL_LOCK;
  s = alloca(size);
  AT_KERNEL_UNLOCK;
  if (!s) {
    perror("at_alloca");
    exit (1);
  }
  return (s);
}




