/*									tab:8
 *
 * ep.h - Endpoint description for the Lanai AM layer.
 *
 * 
 * "Copyright (c) 1995 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 		Richard P. Martin and Lok T. Liu
 * Version:		65
 * Creation Date:	Sun Feb  5 21:46:21 1995
 * Filename:		ep.h
 * History:             
 */

/* all test structs assume that shorts don't get packed */

#define DEPTH      4  /* was 8 */
#define MAX_REMOTE 16 /* was 32 */
#define MAX_HOPS   8

#define IN_Q_SIZE  (2*(DEPTH*MAX_REMOTE))
#define OUT_Q_SIZE (2*(DEPTH*MAX_REMOTE))

#define IN_SIZE_MASK (IN_Q_SIZE-1)    /* QUEUES MUST BE POWERS OF 2 */
#define OUT_SIZE_MASK (OUT_Q_SIZE-1)

#define LAM_PKT_WORDS 512           /* packet size in words */
#define LAM_PKT_SHORTS (LAM_PKT_WORDS*2) /* in shorts */

#define MAX_IT_PERIOD 0x7fff 
/* we have two defs foe what's inside a descriptor, one for the LANai
 * and one for the host. 
 * both should end up being the same size 
 */

#ifdef LANAI

#define OP_MASK   0xf
#define THREAD    0x3
#define SHORT     0x1
#define BULK      0x2
#define BAD_CRC   0x4

#define AM_TYPE_MASK 0xFF00
#define AM_TYPE      0x0100

#define DEST_MASK 0x00ff

typedef struct bulk_data {
  short la_hi;     /* local I/O virtual address to start xfer */
  short la_low;    /* has to be 32-bit aligned                */
  short len_hi;    /* length of xfer in bytes                 */
  short len_low;
  short data[8];  /*  bytes of payload                 */

} bulk_t;

typedef struct desc {
  short dest;    /* or source for input possible timeout field? */    
  short op;      /* own bit and op-code set */
  union {
    short s_data[12];  /* six 32 bit-words of payload  */
    bulk_t b_hdr;
  } hdr;
  short crc;    /* byte for tail flit for CRC check */
  short junk; 
} desc_t ;

typedef struct bulk_d {
  short hi;  /* incomming pointer to bulk data xfer */
  short low;
}  bulk_p_t;

/* host uses int for better access */
#else 

#define AM_TYPE_MASK 0xFF000000
#define AM_TYPE      0x01000000

#define DEST_MASK 0x00ff0000
#define DEST_SHIFT 0x10

#define OP_MASK   0x0000000f
#define THREAD    0x00000003
#define SHORT     0x00000001     /* small packet */
#define BULK      0x00000002     /* bulk xfer    */
#define BAD_CRC   0x00000004

#define REQ_MASK 0x00000010     /* request/reply type */
#define REQUEST  0x00000000     
#define REPLY    0x00000010

#define SEQ_MASK 0x0000ff00     /* request sequence number */
#define SEQ_SHIFT 0x8

typedef struct bulk_data {
  volatile int *la;           /* local pointer to data */
  volatile int len;           /* length in bytes       */
  volatile int data[4];       /* 5 words of payload */
} bulk_t;

/* a descriptor consists of an owned bit, opcode and dest 
 * for small xfers, we get 7 words of payload 
 * for bulk data, we use two words for the src and len and get
 * get 5 words of payload 
 */

typedef struct desc {
  volatile int op_dest;
  union {
    volatile int s_data[6];      /* small payload */
    volatile bulk_t b_hdr;       /* bulk header   */
  } hdr ;
  int crc;
} desc_t ;

typedef struct bulk_d {
  volatile unsigned int *hi;  /* incomming pointer to bulk data xfer */
}  bulk_p_t;

#endif /* host */

/* a route is a length followed by at most 8 hops */
typedef struct route {
  volatile unsigned short len;
  volatile unsigned short pad;  
  volatile unsigned short path[4];
} route_t;

/* private data associated with each endpoint */
typedef struct private_ep {
  volatile unsigned short src;
} private_ep_t;

/* Each endpoint is an 8k block with the following structure */
typedef struct ep {

  bulk_p_t bulk_d_p[IN_Q_SIZE];

  /* request descriptors */
  desc_t out[OUT_Q_SIZE];

  /* reply descriptors   */
  desc_t in[IN_Q_SIZE];

} ep_t ; 


