/*									tab:8
 *
 * lcp.c - Lanai Control Program for the Lanai AM layer.
 *
 * 
 * "Copyright (c) 1995 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 		Richard P. Martin and Lok T. Liu
 * Version:		49
 * Creation Date:	Sun Feb  5 21:46:21 1995
 * Filename:		lcp.c
 * History:             
 */

#include "lanai_def.h"
#include "ifc.h"
#include "ep.h"

/*********/
#define TIME_STAMP(x) asm("mov #-20,r15\n \t mov @r15,%0" : "=m" (x))

#define HOST_DMA_FREE 0x0
#define HOST_DMA_RECV 0x1
#define HOST_DMA_SEND 0x2

/* one endpoint for now */
ep_t endpt;
private_ep_t p_endpt;

route_t routes[MAX_REMOTE]; /* max 32 hosts for now */

/* staging areas  -- 4K each + descriptor + tail flit */
short snd_buf0[LAM_PKT_SHORTS+sizeof(desc_t)+2];
short snd_buf1[LAM_PKT_SHORTS+sizeof(desc_t)+2];

short rcv_buf0[LAM_PKT_SHORTS+sizeof(desc_t)+2];

short time_start,time_stop;

volatile short debug_event;
short *l,*h;
short dropped, bad_type;

volatile int host_DMA_mutex;
volatile int prefetch; 
short *prefetch_buf, *send_buf;

short trace_num;
short trace_buf[16];
#define TRACE_POINT(x) {trace_buf[trace_num]=x ; trace_num=(trace_num+1)%16;}

short pc_trace_num;
short pc_trace_buf[4];
#define PC_TRACE_POINT(x) {pc_trace_buf[pc_trace_num]=x ; \
                           pc_trace_num=(pc_trace_num+1)%4;}
volatile short send_trace_num;
volatile short send_trace_buf[8];
#ifdef DEBUG
#define SEND_TRACE_POINT(x) {send_trace_buf[send_trace_num]=x ; \
                             send_trace_num=(send_trace_num+1)%8;}
#else 
#define  SEND_TRACE_POINT(x) 
#endif 

volatile int send_count;
volatile int recv_count;
volatile int crc_errors; 

volatile int recv_ticks, send_ticks, recv_count, send_count;

volatile int BOGUS_SH;

/***************************************************************/

break_pt(int pc) {
 debug_event=pc;
 TRACE_POINT(0x55);
 TRACE_POINT(APC);
 while (!H_DEBUG_GO);  
 H_DEBUG_GO = 0;
}

void
recv_thread() {
  register desc_t *in_d; /* current and next input descriptors */
  register int in_num;  /* input descriptor number */

  register unsigned short recv_src,recv_op;
  register unsigned short good_crc;

  in_num = 0;
  in_d = &(endpt.in[0]);

  while (1) {
    /* try to receive anything, then send */

#ifdef DEBUG
    recv_ticks++;
#endif 
    
    if (AM_INTR == 1) {
      {
	register int temp;
	temp = (in_num == 0) ? IN_SIZE_MASK : in_num-1;
	if ( (endpt.in[temp].op&OP_MASK) != 0)
	  INT =1;
      }
    }

    if (ISR & LONG_RDY_BIT ) {

#ifdef DEBUG
    recv_count++;
#endif 

      recv_src = RH;   /* can get away with 2 RH's since a packet > 4bytes*/
      recv_op = RH;

    if ((recv_src & AM_TYPE_MASK) != AM_TYPE) {
      bad_type++;
      goto drop;
    }
      /* check if current descriptor is free */
      if ( (in_d->op&OP_MASK) == 0) {

	/*MAKE SURE CRC BYTE DOESN'T COUNT!*/
	RMP = (int ) &(in_d->hdr.s_data[0]) ;
	RML = (int ) &(in_d->crc) ; 
	/* switch to send thread while waiting for header */
	/* if we get a small packet, both will be set, but on a */
        /* bulk packet, then only buff int gets set */

	while (!(ISR & (RECV_INT_BIT|BUFF_INT_BIT)))
	  RESUME_USER;

	in_d->dest = recv_src ;

	/* see if It's a small packet */
	/* if ( (recv_op&OP_MASK) == SHORT) {  SHORT or THREADS */
	if ( (recv_op&SHORT)) {

	  if ( (in_d->crc&0xff00) == 0 ) { /* good tail flit */
	    in_d->op = recv_op;	 /* end nextbuf free*/
	  }
	  else {
	    in_d->op = recv_op|BAD_CRC;
	    crc_errors++;
	  }
	}
	else {			/* a bulk packet */
	  /* try to suck in whole body */
          /* because the tail flit is appended to the packet, and counts 
           * as part of the length, the first 4 bytes got left in the CRC
           * Part of the header....
           * hack is to copy them here and set the pointer two down ..
           * Grrr.
           */

	  *rcv_buf0 = in_d->crc;
	  *(rcv_buf0+1) = in_d->junk;

	  RMP = (int) &(rcv_buf0[2]);
	  RML = (int) &(rcv_buf0[LAM_PKT_SHORTS+2]) ;
	  /* switch to send thread while waiting for msg to complete */	 
	  while (!(ISR & RECV_INT_BIT))
	    RESUME_USER;

	  /* check CRC */
	  good_crc =( (((*(((unsigned * ) RMP) - 2)) >> 8) & 0xff) == 0);

	  if (good_crc) {	/* good */
	    /* while DMA is busy, switch to send thread */
	    while (host_DMA_mutex != HOST_DMA_FREE)
	      RESUME_USER;

	    host_DMA_mutex = HOST_DMA_RECV;	/* mark host DMA as busy */
	    /* move data into host */
	    EARM = endpt.bulk_d_p[in_num].hi;
	    EARL = endpt.bulk_d_p[in_num].low;
	    LAR = (int) rcv_buf0;
	    DMA_DIR = L2E;
	    DMA_CTR = (in_d->hdr.b_hdr.len_low)>>1;
	    /* while DMA is not done, switch to send thread */

	    while (!(ISR & DMA_INT_BIT)) {
	      RESUME_USER;  
	    }
	    host_DMA_mutex = HOST_DMA_FREE;	/* mark host DMA as free */
	    in_d->op = recv_op;
	  }

	  /* bad CRC */
	  else {
	    in_d->op = recv_op|BAD_CRC;
	    crc_errors++;
	  }
	} /* end bulk */
	in_num = (in_num+1) & IN_SIZE_MASK;
	in_d = &(endpt.in[in_num]);
      }
      /* if the input buffer was full, don't advance the pointer, we */
      /* just drop everything on the floor */
      else { 
	dropped++;
      drop:
	RMP = (int) &(rcv_buf0[0]);
	RML = (int) &(rcv_buf0[LAM_PKT_SHORTS+sizeof(desc_t)+2]) ;
	/* switch to send thread while waiting for msg to complete */
	while (!(ISR & RECV_INT_BIT))
	  RESUME_USER; 

      }
    } /* anything to receive */
    RESUME_USER;		/* switch to send thread */
  }
} /* recv_thread */

void 
send_thread()
{

  register desc_t *out_d,*out_d_low,*out_d_high;
  register desc_t *out_d_next;  /* current output descriptor */
  register int dest; 

  out_d = &(endpt.out[0]);
  out_d_next =  &(endpt.out[1]);

  out_d_low =  &(endpt.out[0]);
  out_d_high = &(endpt.out[OUT_SIZE_MASK]);

  prefetch =0;
  send_buf = snd_buf0;
  prefetch_buf = snd_buf1;

  while (1) {

  send_top: /* switch to receive thread while no work */

#ifdef DEBUG 
    send_ticks++;
#endif 

    /* if host posts a send request */
    /* if ( (out_d->op&OP_MASK) == SHORT ) {    SHORT or THREADS */
    if((out_d->op&SHORT)) {

#ifdef DEBUG 
    send_count++;
#endif 

      while (!(ISR & SEND_RDY_BIT)) {
	RESUME_SYSTEM;
      }


    dest = out_d->dest & DEST_MASK;
    
    switch (routes[dest].len) {
    case 0:
       /* special case refects to self */
       goto skip_wait;
     case 1:
       SB = (routes[dest].path[0]);  
       goto skip_wait;
     case 2:
       SH = (routes[dest].path[0]);
       goto skip_wait;
     case 3:
       SA = 1;
       SMP = (int)(routes[dest].path);
       SML = (int)(routes[dest].path);
       break;
     case 4:
       SMP = (int)(routes[dest].path);
       SML = (int)(routes[dest].path); 
       break; 
     case 5:
       SA = 3;
       SMP = (int)(routes[dest].path);
       SML = (int)(routes[dest].path+2);
       break;
     case 6:
       SA = 2;
       SMP = (int)(routes[dest].path);
       SML = (int)(routes[dest].path+2);
       break;
     case 7:
       SA = 1;
       SMP = (int)(routes[dest].path);
       SML = (int)(routes[dest].path+2);
       break;
     case 8:
       SMP = (int)(routes[dest].path);
       SML = (int)(routes[dest].path+2);
       break;
     default:
       break; 
     }

    while (!(ISR & SEND_INT_BIT)) {
      RESUME_SYSTEM; 
    } 
   
  skip_wait:
      /* overwrite dest with source word */
      out_d->dest = p_endpt.src|AM_TYPE;
      SMP = (int) &(out_d->dest);
      SMLT = (int) &(out_d->hdr.s_data[10]); 

      /* switch to receive thread while waiting to send out header */
      while (!(ISR & SEND_INT_BIT)) {
	RESUME_SYSTEM; 
      }

      /* clear op-code */
      out_d->op = out_d->op & (~OP_MASK);

      /* inc and mod */
      out_d = out_d_next;
      if (out_d_next == out_d_high) 
	out_d_next = out_d_low;
      else 
	out_d_next++;

      RESUME_SYSTEM;
      SEND_TRACE_POINT(0xa);
      goto send_top;

    } /* short message */

    /* this bulk packet had it's data prefetch from last time */
    if (prefetch) { 

#ifdef DEBUG 
    send_count++;
#endif 

    send_prefetch:
      /* swap prefetch and send buffers */
      {  short *temp;
	 temp = send_buf;
	 send_buf = prefetch_buf;
	 prefetch_buf = temp;
	 prefetch =0;
       }

      while (!(ISR & DMA_INT_BIT)) {
	RESUME_SYSTEM;
      }

      host_DMA_mutex = HOST_DMA_FREE;    /* mark host DMA as free */
      RESUME_SYSTEM;        /* give receive a shot at DMA */
      
      /* send header out to reduce latency as much as possible */
      while (!(ISR & SEND_RDY_BIT)) {
	RESUME_SYSTEM;
      }


    dest = out_d->dest & DEST_MASK;
    
    switch (routes[dest].len) {
    case 0:
       goto skip_wait2;  /* special case refects to self */
     case 1:
       SB = (routes[dest].path[0]);  
       goto skip_wait2;
     case 2:
       SH = (routes[dest].path[0]);
       goto skip_wait2;
     case 3:
       SA = 1;
       SMP = (int)(routes[dest].path);
       SML = (int)(routes[dest].path);
       break;
     case 4:
       SMP = (int)(routes[dest].path);
       SML = (int)(routes[dest].path); 
       break; 
     case 5:
       SA = 3;
       SMP = (int)(routes[dest].path);
       SML = (int)(routes[dest].path+2);
       break;
     case 6:
       SA = 2;
       SMP = (int)(routes[dest].path);
       SML = (int)(routes[dest].path+2);
       break;
     case 7:
       SA = 1;
       SMP = (int)(routes[dest].path);
       SML = (int)(routes[dest].path+2);
       break;
     case 8:
       SMP = (int)(routes[dest].path);
       SML = (int)(routes[dest].path+2);
       break;
     default:
       break; 
     }

    while (!(ISR & SEND_INT_BIT)) {
      RESUME_SYSTEM; 
    } 
   
  skip_wait2:
      /* overwrite dest with source word */
      out_d->dest = p_endpt.src|AM_TYPE;
      SMP = (int) &(out_d->dest);
      SML = (int) &(out_d->hdr.s_data[10]); 
      
      /* try to get next packet */
      if ( (out_d_next->op&OP_MASK) == BULK) { 
	/* if we can't grab the mutex abort */
	/* the prefetch, since receive has it */
	if  (host_DMA_mutex == HOST_DMA_FREE) { 
	  host_DMA_mutex = HOST_DMA_SEND;   /* mark host DMA as busy */
	  /* move data from host to LANai memory */
	  EARM = out_d_next->hdr.b_hdr.la_hi;
	  EARL = out_d_next->hdr.b_hdr.la_low;
	  LAR = (int) prefetch_buf;
	  DMA_DIR = E2L;
	  DMA_CTR = (out_d_next->hdr.b_hdr.len_low)>>1;
	  prefetch = 1;
	} /* we got the DMA mutex */
      } /* try to prefetch the next bulk packet */

      /* wait for header */
      while (!(ISR & SEND_INT_BIT)) {
	RESUME_SYSTEM; 
      }

      SMP = (int) send_buf;
      SMLT = (int) &(send_buf[(out_d->hdr.b_hdr.len_low)>>1]);

      /* switch to receive thread while DMA is not done */
      if ( host_DMA_mutex == HOST_DMA_FREE) {
	  SEND_TRACE_POINT(0x1);
	  while (!(ISR & SEND_INT_BIT))
	    RESUME_SYSTEM;
	}
      else {
	while (!(ISR & SEND_INT_BIT)) {
	  SEND_TRACE_POINT(0x2);
	  if (ISR & DMA_INT_BIT) 
	    host_DMA_mutex = HOST_DMA_FREE;
	  RESUME_SYSTEM;
	}
      }
      /* clear op-code */
      out_d->op = out_d->op & (~OP_MASK);

      /* inc and mod */
      out_d = out_d_next;
      if (out_d_next == out_d_high) 
	out_d_next = out_d_low;
      else 
	out_d_next++;

      RESUME_SYSTEM;
      SEND_TRACE_POINT(0x3);
      goto send_top;
    } /* end if prefetch */
    
    /* first bulk packet, start the prefetch */
    if ( (out_d->op&OP_MASK) == BULK ) {  

      while (host_DMA_mutex != HOST_DMA_FREE) {
	RESUME_SYSTEM;
      }

      host_DMA_mutex = HOST_DMA_SEND;   /* mark host DMA as busy */
      /* move data from host to LANai memory */
      EARM = out_d->hdr.b_hdr.la_hi;
      EARL = out_d->hdr.b_hdr.la_low;
      LAR = (int) prefetch_buf;
      DMA_DIR = E2L;
      DMA_CTR = (out_d->hdr.b_hdr.len_low)>>1;

      prefetch = 1;
      SEND_TRACE_POINT(0x5);
      goto send_prefetch;
    }

    RESUME_SYSTEM;
    SEND_TRACE_POINT(0x7);
  } /* end main loop */

}

extern "C" void set_user_context();

int
main()
{
  int j;

  if (ISR & FRES_INT_BIT) {
    TRACE_POINT(0x67);
    break_pt(PC);
  }

  APC = (int) set_user_context;

  RESUME_USER;
  APC = (int) send_thread;

  send_ticks = recv_ticks = send_count = recv_count = crc_errors = 0;

  debug_event =0;
  host_DMA_mutex = 0;    /* 0 = DMA is free */
  dropped=bad_type=0;

  /* this code assumes host always zeros lanai mem.     */
  /* doesn't work if we get hit with a forward reset   */

  clear_FRES_INT_BIT();
  DMA_STS = 0x7; /* H_DMA_STS; works for sparc 10 and 20 */
  STOP_GO = 1;
  CRC = 1;


  /* reset always generates two garbage tail flits ... */
  /* eat two garbage tail flits in the interface on reset */
  for (j = 0; j < 2; j++){
    while (!(ISR & TAIL_INT_BIT))
      {
	int i;
	while (!(ISR & BYTE_RDY_BIT));
	i = RB;
      }
    clear_TAIL_INT_BIT();
  }

  trace_num= pc_trace_num = send_trace_num=send_count= recv_count=0;

  ROUTE_TAB =  (int) &(routes[0]);
  IN_BASE   = (int )  &(endpt.in[0]);
  OUT_BASE = (int)  &(endpt.out[0]);
  BULK_PTR  = (int) &(endpt.bulk_d_p[0]);
  P_END_PT = (int) &(p_endpt);

  IMR = 0;

  /* tell host we are ready */
  L_INIT_DONE =1;

  /* host must set up recv queue*/
  TRACE_POINT(0x01);

  /* host tells us it's ready */
  while (!H_INIT_DONE);  

  TRACE_POINT(0x02);
  TRACE_POINT(0x03);
  recv_thread();

} /* end main */




