/* 
 * Copyright (C) International Computer Science Institute, 1995/1996.  COPYRIGHT
 * NOTICE: This code is provided "AS IS" WITHOUT ANY WARRANTY and is subject
 * to the terms of the GNU LIBRARY GENERAL PUBLIC LICENSE contained in
 * the file "Doc/LGPL" of the Sather distribution.  The license is also
 * available from ICSI, 1947 Center St., Suite 600, Berkeley CA 94704, USA.
 * ---> Please email bugs to "psather-bugs@icsi.berkeley.edu". 
 *
 * Local memory allocation 
 *
 * Version 1.0 (released for 1.0.9) by Claudio Fleiner (fleiner@icsi.berkeley.edu)
 */
#ifndef _MEMORY_H_
#define _MEMORY_H_

#ifdef AT_THREADS
#include "at.h"
#else
#include <malloc.h>
#endif

/* Try zeroing memory everywhere */

#ifndef PSATHER1
#define MALLOC(r,s)  	do { (void*)r=(void*)malloc((s));if(r==NULL || ((long)r&CLUSTER_MASK)) {fprintf(stderr,"%s:%d: out of memory (malloc(%ld))\n",__FILE__,__LINE__,(long)(s));abort();}else{memset((void *)r,0,s);} } while(0)
#define FREE(s)  	do { if(s==NULL) {fprintf(stderr,"%s:%d: free(NULL)\n",__FILE__,__LINE__);abort();}free(s); } while(0)
#define REALLOC(r,a,b) 	do { (void*)r=(void*)realloc((a),(b));if(r==NULL || ((long)r&CLUSTER_MASK)) {fprintf(stderr,"%s:%d: out of memory (realloc(%ld))\n",__FILE__,__LINE__,(long)(b));abort();} } while(0)
#define CALLOC(r,a,b)	do { (void*)r=(void*)calloc((a),(b));if(r==NULL || ((long)r&CLUSTER_MASK)) {fprintf(stderr,"%s:%d: out of memory (calloc(%ld,%ld))\n",__FILE__,__LINE__,(long)(a),(long)(b));abort();}else{memset((void *)r,0,a*b);} } while(0)
#else
#define MALLOC(r,s)  	do { (void*)r=(void*)malloc((s));if(r==NULL) {fprintf(stderr,"%s:%d: out of memory (malloc(%ld))\n",__FILE__,__LINE__,(long)(s));abort();}else{memset((void *)r,0,s);} } while(0)
#define FREE(s)  	do { if(s==NULL) {fprintf(stderr,"%s:%d: free(NULL)\n",__FILE__,__LINE__);abort();}free(s); } while(0)
#define REALLOC(r,a,b) 	do { (void*)r=(void*)realloc((a),(b));if(r==NULL) {fprintf(stderr,"%s:%d: out of memory (realloc(%ld))\n",__FILE__,__LINE__,(long)(b));abort();} } while(0)
#define CALLOC(r,a,b)	do { (void*)r=(void*)calloc((a),(b));if(r==NULL) {fprintf(stderr,"%s:%d: out of memory (calloc(%ld,%ld))\n",__FILE__,__LINE__,(long)(a),(long)(b));abort();} else{memset((void *)r,0,a*b);}} while(0)
#endif

#endif
